/*
  ProgEx32  -- A More Memory Efficient String Class.  C++ Compilation/Linking

  As mentioned in ProgEx31, we are going to try to alter our String Class so it
  isn't always thrashing around in memory allocating/deallocating memory blocks.
  To do that we'll add one more private data member to our class which will keep
  track of the current maximum number of characters the String can hold.  We'll
  name the data member 'iAllowableCharacterCount', and it will be one less than
  whatever the present memory allocation is.  Another change is that we'll
  define an equate named MINIMUM_ALLOCATION.  Unlike in ProgEx31's Uninitialized
  Constructor where no memory allocations occured, in this example a declaration
  like so...

  String s1;

  will cause the Constructor to allocate an amount of memory equal to the above
  equate, which in this program has been set to 8 bytes.  Yet another change is
  that whenever the various member functions determine that not enough memory is
  indicated by this->iAllowableCharacterCount, rather than only re-allocating
  the needed amount, the needed amount will be multiplied by another equate
  named EXPANSION_FACTOR so as to anticipate future String growth.  In this
  program EXPANSION_FACTOR has been set to 2.  Below is an exact copy of our
  String Declaration header showing what our String Class now looks like...

  //Strings.h                         Class Declaration For String Class So Far
  #if !defined(STRINGS_H)             //This is an inclusion guard to guard against other
  #define STRINGS_H                   //files that may have previously included this file.
  #define EXPANSION_FACTOR      2     //This is the factor to anticipate future space needs
  #define MINIMUM_ALLOCATION    8     //Allocations will be at least this number

  class String
  {
   public:
   String();                         //Uninitialized Constructor - Does nothing
   String(const char);               //Constructor Initializes String With char
   String(const char*);              //Constructor Initializes String With char*
   String(const String&);            //Constructor Initializes String With Another String (Copy Constructor)
   String& operator=(const char);    //Operator= for assigning char to String
   String& operator=(const char*);   //Operator= for assigning literal or char* to String
   String& operator=(const String&); //Operator= for assigning 2nd String to 1st
   String& operator+(const char);    //Operator+ for adding char to String
   String& operator+(const char*);   //Operator+ for adding A Null Terminated Ascii String Array
   String& operator+(const String&); //Operator+ for adding Another String To The String
   int LenStr(void);                 //Returns the length of what's pointed to by pStrBuffer
   char* lpStr(void);                //Returns Address of String data Member  >> this->lpStr()
   ~String();                        //String Destructor

   private:
   char* pStrBuffer;                 //This char* member holds the address of an allocated String
   int   iAllowableCharacterCount;   //Max number of chars that will fit in presently allocated buffer
  };
  #endif

  To see the improvement, lets run the example we had in the last ProgEx where
  we concatenated a few letters of the English alphabet together, except this
  time we'll do the whole alphabet.  If we would have done that in the last
  example (ProgEx31) we would have had 26 allocations and 26 releases.  You can
  find out for yourself in the output below how the changes we made affects this
  here.  My comments on the code will follow the example and the output
  immediately beneath...
*/


#include <stdio.h>   //The String declaration below 'String s1' will cause an 8
#include "Strings.h" //byte memory allocation.  The logic in the for loop will
                     //cause calls to the overloaded operator+ member, and that
int main(void)       //member function will check the present allowable char
{                    //count against the present count of chars accumulated in
 String s1;          //the String.  Only when the char count reaches 7 will the
                     //function re-allocate memory and release the original 8
 for(unsigned int i=65; i<91; i++)  //bytes.  However, it will double the
 {                                  //present maximum char count and round to a
     s1=s1+i;                       //16 byte paragraph boundary when it
     printf("\n%s\n",s1.lpStr());   //calculates its new memory allocation size.
 }                   //Therefore, the original 8 byte allocation will be doubled
 getchar();          //to 16 bytes. The loop will continue running, accumulating
                     //chars, and checking bytes until it reaches 15 chars, and
 return 0;           //then it will re-allocate again - but this last time for
}                    //32 bytes.  Therefore, there will only need to be two
                     //two re-allocations after the original 8 byte allocation
                     //to finish the alphabet-not 26 like in our first example!


/*  --Output--

(You'll Need About 325 Lines In Your Console Window Buffer)

Entering Uninitialized String Constructor
  pStrBuffer = 4079520
  this->iAllowableCharacterCount = 7
Leaving Uninitialized String Constructor


Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 0
  this->iAllowableCharacterCount = 7
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = A
Early Exit Leaving String::operator=(const String& strRight)

A

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 1
  this->iAllowableCharacterCount = 7
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = AB
Early Exit Leaving String::operator=(const String& strRight)

AB

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 2
  this->iAllowableCharacterCount = 7
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABC
Early Exit Leaving String::operator=(const String& strRight)

ABC

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 3
  this->iAllowableCharacterCount = 7
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCD
Early Exit Leaving String::operator=(const String& strRight)

ABCD

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 4
  this->iAllowableCharacterCount = 7
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDE
Early Exit Leaving String::operator=(const String& strRight)

ABCDE

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 5
  this->iAllowableCharacterCount = 7
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEF
Early Exit Leaving String::operator=(const String& strRight)

ABCDEF

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 6
  this->iAllowableCharacterCount = 7
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFG
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFG

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 7
  this->iAllowableCharacterCount = 7
  !!!!! Memory Wasn't Adequate.  We Need To Re-Allocate!
  this->iAllowableCharacterCount = 15
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGH
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGH

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 8
  this->iAllowableCharacterCount = 15
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGHI
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGHI

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 9
  this->iAllowableCharacterCount = 15
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGHIJ
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGHIJ

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 10
  this->iAllowableCharacterCount = 15
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGHIJK
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGHIJK

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 11
  this->iAllowableCharacterCount = 15
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGHIJKL
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGHIJKL

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 12
  this->iAllowableCharacterCount = 15
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGHIJKLM
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGHIJKLM

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 13
  this->iAllowableCharacterCount = 15
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGHIJKLMN
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGHIJKLMN

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 14
  this->iAllowableCharacterCount = 15
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGHIJKLMNO
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGHIJKLMNO

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 15
  this->iAllowableCharacterCount = 15
  !!!!! Memory Wasn't Adequate.  We Need To Re-Allocate!
  this->iAllowableCharacterCount = 31
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGHIJKLMNOP
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGHIJKLMNOP

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 16
  this->iAllowableCharacterCount = 31
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGHIJKLMNOPQ
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGHIJKLMNOPQ

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 17
  this->iAllowableCharacterCount = 31
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGHIJKLMNOPQR
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGHIJKLMNOPQR

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 18
  this->iAllowableCharacterCount = 31
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGHIJKLMNOPQRS
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGHIJKLMNOPQRS

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 19
  this->iAllowableCharacterCount = 31
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGHIJKLMNOPQRST
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGHIJKLMNOPQRST

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 20
  this->iAllowableCharacterCount = 31
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGHIJKLMNOPQRSTU
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGHIJKLMNOPQRSTU

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 21
  this->iAllowableCharacterCount = 31
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGHIJKLMNOPQRSTUV
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGHIJKLMNOPQRSTUV

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 22
  this->iAllowableCharacterCount = 31
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGHIJKLMNOPQRSTUVW
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGHIJKLMNOPQRSTUVW

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 23
  this->iAllowableCharacterCount = 31
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGHIJKLMNOPQRSTUVWX
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGHIJKLMNOPQRSTUVWX

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 24
  this->iAllowableCharacterCount = 31
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGHIJKLMNOPQRSTUVWXY
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGHIJKLMNOPQRSTUVWXY

Entering String& String::operator+(const char ch) For Adding A char To A String
  this->LenStr()                 = 25
  this->iAllowableCharacterCount = 31
  Memory Was Adequate.  Didn't Need To Re-Allocate!
Leaving String& String::operator+(const char ch) For Adding A char To A String

Entering operator=(const String& strRight) Assigns Another String To A String
  strRight.pStrBuffer = ABCDEFGHIJKLMNOPQRSTUVWXYZ
Early Exit Leaving String::operator=(const String& strRight)

ABCDEFGHIJKLMNOPQRSTUVWXYZ
*/


/*  ....continued from above

  I'd call this a considerable improvement.  We've cut down the memory alloc-
  ations from 16 down to 3.  Obviously, if MINIMUM_ALLOCATION was set high
  enough the concatenation of the alphabet wouldn't take any more re-allocations
  at all.  However, I suppose this could be construed as cheating!  In any case
  achieving this improvement only made the code a small amount more complicated.
  The way things are now all the member functions that either assign Strings or
  add to them (concatenating operator+ members) will have to compare the
  parameter to the function call with the amount of memory presently held as
  determined by the new iAllowableCharacterCount private member.  If enough
  memory is being held, rather than a re-allocation/de-allocation sequence the
  parameter is simply copied to the existing memory.  If enough memory isn't
  present then a new allocation will be made after multiplying the present need
  by EXPANSION_FACTOR.  For example, allocations could follow a sequence of 8,
  16, 32, 64, 128.... etc.

  String& String::operator+(const char ch)
  {
   int iLen,iNewSize;
   char* pNew;

   iLen=strlen(this->pStrBuffer);
   if(iLen<this->iAllowableCharacterCount)
   {
      this->pStrBuffer[iLen]=ch;
      this->pStrBuffer[iLen+1]='\0';
   }
   else
   {
      iNewSize=((this->iAllowableCharacterCount*EXPANSION_FACTOR)/16+1)*16;
      pNew=new char[iNewSize];
      this->iAllowableCharacterCount=iNewSize-1;
      strcpy(pNew,this->pStrBuffer);
      delete [] this->pStrBuffer;
      this->pStrBuffer=pNew;
      this->pStrBuffer[iLen]=ch;
      this->pStrBuffer[iLen+1]='\0';
   }

   return *this;
  }

  This term...

  iNewSize=((this->iAllowableCharacterCount*EXPANSION_FACTOR)/16+1)*16;

  has an integer division within it that in C/C++ languages ends up usually in
  an integer truncation.  For example, when our String above reached...

  ABCDEFG

  ...and we had this situation...

  this->LenStr()                 = 7
  this->iAllowableCharacterCount = 7

  ...the formula above would have resolved to this...

  iNewSize=((7*2)/16+1)*16;

  ...which due to integer truncation resolves to this...

  iNewSize = 16;
*/
